#include "arm_control.h"
#include "exceptions.h"
#include <iostream>
#include <fstream>
#include <unistd.h>
#include <sys/select.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <errno.h>
#include <math.h>
#include <linux/joystick.h>
#include "xml/arm_motions_cfg_file.hxx"

void button_callback(void *param,unsigned int button_id,bool level)
{
  CArmControl *arm=(CArmControl *)param;

  try{
    switch(button_id)
    {
      case 4: if(level)
                arm->increase_max_speed();
              break; 
      case 6: if(level)
                arm->decrease_max_speed();
              break; 
      case 8: if(level)
                arm->enable_torque(false); 
              else
                arm->enable_torque(true);
              break;
      case 10: if(!level)
               {
                 // button rear left 1 released -> keep position
                 arm->stop();
               }
               break;
      case 12: if(level)
               {
                 if(arm->get_button_state(11))
                   arm->update_pose(0);// update circle pose
                 else
                   arm->execute_pose(0);// execute circle pose
               }
               break; 
      case 13: if(level)
               {
                 if(arm->get_button_state(11))
                   arm->update_pose(1);// update circle pose
                 else
                   arm->execute_pose(1);// execute circle pose
               }
               break; 
      case 14: if(level)
               {
                 if(arm->get_button_state(11))
                   arm->update_pose(2);// update circle pose
                 else
                   arm->execute_pose(2);// execute circle pose
               }
               break; 
      case 15: if(level)
               {
                 if(arm->get_button_state(11))
                   arm->update_pose(3);// update circle pose
                 else
                   arm->execute_pose(3);// execute circle pose
               }
               break; 
    }
  }catch(CException &e){
    std::cout << e.what() << std::endl;
  }
}

void axis_callback(void *param,unsigned int axis_id, float value)
{
  CArmControl *arm=(CArmControl *)param;

  try{
    switch(axis_id)
    {
      case 0: if(arm->get_button_state(10))
                arm->move(PAN,value);
              break;
      case 1: if(arm->get_button_state(10))
                arm->move(TILT,value);
              break;
      case 2: if(arm->get_button_state(10))
                arm->move(ROLL,value);
              break;
    }
  }catch(CException &e){
    std::cout << e.what() << std::endl;
  }
}

CArmControl::CArmControl(const std::string &joy_device,std::string &dyn_serial, unsigned int baudrate,unsigned char pan_id, unsigned char tilt_id, unsigned char roll_id) : joy("arm_control_jpy")
{
  std::string name;

  this->exit_flag=false; 
  // detect and initialize servos
  this->dyn_server=CDynamixelServerFTDI::instance();
  if(this->dyn_server->get_num_buses()>0)
  {
    this->dyn_server->config_bus(dyn_serial,baudrate);
    sleep(1);
    this->motors.resize(3);
    name="arm_control_pan";
    this->motors[PAN] = new CDynamixelMotor(name,this->dyn_server,pan_id);
    this->set_max_angles(PAN,90.0,-90.0);
    this->set_max_torque(PAN,50.0);
    name="arm_control_tilt";
    this->motors[TILT] = new CDynamixelMotor(name,this->dyn_server,tilt_id);
    this->set_max_angles(TILT,90.0,-90.0);
    this->set_max_torque(TILT,50.0);
    name="arm_control_roll";
    this->motors[ROLL] = new CDynamixelMotor(name,this->dyn_server,roll_id);
    this->set_max_angles(ROLL,90.0,-90.0);
    this->set_max_torque(ROLL,50.0);
  }
  else
    throw CException(_HERE_,"No Dynamixel buses available");
  this->stored_poses.resize(4);
  this->max_speed=50.0;
  this->enable_torque(true);
  // open the joystick device
  this->joy.open(joy_device);
  this->current_button_state.resize(this->joy.get_num_buttons());
  for(unsigned int i=0;i<this->joy.get_num_buttons();i++)
    this->current_button_state[i]=this->joy.get_button_state(i);
  // assign callbacks
  this->joy.enable_button_callback(4,button_callback,this);
  this->joy.enable_button_callback(6,button_callback,this);
  this->joy.enable_button_callback(8,button_callback,this);
  this->joy.enable_button_callback(10,button_callback,this);
  this->joy.enable_button_callback(12,button_callback,this);
  this->joy.enable_button_callback(13,button_callback,this);
  this->joy.enable_button_callback(14,button_callback,this);
  this->joy.enable_button_callback(15,button_callback,this);
  this->joy.enable_position_change_callback(0,axis_callback,this);
  this->joy.enable_position_change_callback(1,axis_callback,this);
  this->joy.enable_position_change_callback(2,axis_callback,this);
  // create threads and events
  this->thread_server=CThreadServer::instance();
  this->position_monitor_thread_id="arm_control_position_monitor_thread";
  this->thread_server->create_thread(this->position_monitor_thread_id);
  this->thread_server->attach_thread(this->position_monitor_thread_id,this->position_monitor_thread,this);
  this->event_server=CEventServer::instance();
  this->finish_thread_event_id="arm_control_finish_thread_event";
  this->event_server->create_event(this->finish_thread_event_id);
  this->thread_server->start_thread(this->position_monitor_thread_id);
}

void *CArmControl::position_monitor_thread(void *param)
{
  CArmControl *arm=(CArmControl *)param;
  double current_pos;
  bool end=false;

  while(!end)
  {
    if(arm->event_server->event_is_set(arm->finish_thread_event_id))
      end=true;
    else
    { 
      try{
        if(arm->get_button_state(10))
        {
          std::cout << "thread" << std::endl;
          current_pos=arm->motors[PAN]->get_current_angle();
          if(current_pos>arm->max_angles.pan && arm->target_torques.pan>0.0)
            arm->motors[PAN]->move_absolute_angle(current_pos,50.0);
          else if(current_pos<arm->min_angles.pan && arm->target_torques.pan<0.0)
            arm->motors[PAN]->move_absolute_angle(current_pos,50.0);
          else
            arm->motors[PAN]->move_torque(arm->target_torques.pan);
          current_pos=arm->motors[TILT]->get_current_angle();
          if(current_pos>arm->max_angles.tilt && arm->target_torques.tilt>0.0)
            arm->motors[TILT]->move_absolute_angle(current_pos,50.0);
          else if(current_pos<arm->min_angles.tilt && arm->target_torques.tilt<0.0)
            arm->motors[TILT]->move_absolute_angle(current_pos,50.0);
          else
            arm->motors[TILT]->move_torque(arm->target_torques.tilt);
          current_pos=arm->motors[ROLL]->get_current_angle();
          if(current_pos>arm->max_angles.roll && arm->target_torques.roll>0.0)
            arm->motors[ROLL]->move_absolute_angle(current_pos,50.0);
          else if(current_pos<arm->min_angles.roll && arm->target_torques.roll<0.0)
            arm->motors[ROLL]->move_absolute_angle(current_pos,50.0);
          else
            arm->motors[ROLL]->move_torque(arm->target_torques.roll);
        }
        if(arm->get_button_state(0))
          arm->exit_flag=true;
      }catch(CException &e){
      }
    }
    usleep(30000);
  }

  pthread_exit(NULL);
}

bool CArmControl::get_button_state(unsigned int button_id)
{
  if(button_id>this->current_button_state.size())
    throw CException(_HERE_,"Invalid button ID");
  else
    return this->joy.get_button_state(button_id);
}

bool CArmControl::exit(void)
{
  return this->exit_flag;
}

void CArmControl::set_max_angles(unsigned int axis_id, double max, double min)
{
  if(axis_id==PAN)
  {
    this->max_angles.pan=max;
    this->min_angles.pan=min;
    this->motors[PAN]->set_position_range(min,max);
  }
  else if(axis_id==TILT)
  {
    this->max_angles.tilt=max;
    this->min_angles.tilt=min;
    this->motors[TILT]->set_position_range(min,max);
  }
  else if(axis_id==ROLL)
  {
    this->max_angles.roll=max;
    this->min_angles.roll=min;
    this->motors[ROLL]->set_position_range(min,max);
  }
  else
    throw CException(_HERE_,"Invalid axis ID");
}

void CArmControl::set_max_torque(unsigned int axis_id, double max)
{
  if(max>100.0)
    max=100.0;
  else if(max<0.0)
    max=0.0;

  if(axis_id==PAN)
  {
    this->motors[PAN]->set_max_torque(max);
    this->motors[PAN]->set_limit_torque(max);
    this->max_torques.pan=max;
  }
  else if(axis_id==TILT)
  {
    this->motors[TILT]->set_max_torque(max);
    this->motors[TILT]->set_limit_torque(max);
    this->max_torques.tilt=max;
  }
  else if(axis_id==ROLL)
  {
    this->motors[ROLL]->set_max_torque(max);
    this->motors[ROLL]->set_limit_torque(max);
    this->max_torques.roll=max;
  }
  else
    throw CException(_HERE_,"Invalid axis ID");
}

void CArmControl::enable_torque(bool enable)
{
  if(this->motors.size()!=3)
    throw CException(_HERE_,"Not enough servos avaialble");
  if(this->motors[PAN]!=NULL)
  {
    if(enable)
    {
      std::cout << "enable torque pan" << std::endl;
      this->motors[PAN]->enable();
    }
    else
    {
      std::cout << "disable torque pan" << std::endl;
      this->motors[PAN]->disable();
    }
  }
  else
    throw CException(_HERE_,"PAN motor not properly initialized");
  if(this->motors[TILT]!=NULL)
  {
    if(enable)
    {
      std::cout << "enable torque tilt" << std::endl;
      this->motors[TILT]->enable();
    }
    else
    {
      std::cout << "disable torque tilt" << std::endl;
      this->motors[TILT]->disable();
    }
  }
  else
    throw CException(_HERE_,"TILT motor not properly initialized");
  if(this->motors[ROLL]!=NULL)
  {
    if(enable)
    {
      std::cout << "enable torque roll" << std::endl;
      this->motors[ROLL]->enable();
    }
    else
    {
      std::cout << "disable torque roll" << std::endl;
      this->motors[ROLL]->disable();
    }
  }
  else
    throw CException(_HERE_,"ROLL motor not properly initialized");
}

void CArmControl::move(unsigned int axis_id,double torque)
{
  if(axis_id==PAN)
  {
    std::cout << "move pan" << std::endl;
    this->target_torques.pan=(torque*this->max_torques.pan)/32767.0;
  }
  else if(axis_id==TILT)
  {
    std::cout << "move tilt" << std::endl;
    this->target_torques.tilt=(torque*this->max_torques.tilt)/32767.0;
  }
  else if(axis_id==ROLL)
  {
    std::cout << "move roll" << std::endl;
    this->target_torques.roll=(torque*this->max_torques.roll)/32767.0;
  }
  else
    throw CException(_HERE_,"Invalid axis ID");  
}

void CArmControl::stop(void)
{
  double pos;

  pos=this->motors[PAN]->get_current_angle();
  this->motors[PAN]->move_absolute_angle(pos,50.0);
  pos=this->motors[TILT]->get_current_angle();
  this->motors[TILT]->move_absolute_angle(pos,50.0);
  pos=this->motors[ROLL]->get_current_angle();
  this->motors[ROLL]->move_absolute_angle(pos,50.0);
}

void CArmControl::increase_max_speed(void)
{
  this->max_speed+=10.0;
  if(this->max_speed>100.0)
    this->max_speed=100.0;
  std::cout << "speed: " << this->max_speed << std::endl;
}

void CArmControl::decrease_max_speed(void)
{
  this->max_speed-=10.0;
  if(this->max_speed<10.0)
    this->max_speed=10.0;
  std::cout << "speed: " << this->max_speed << std::endl;
}

void CArmControl::load_poses(const std::string &filename)
{
  arm_motions_t::arm_motion_iterator iterator;
  TJoints new_motion;
  struct stat buffer;
  unsigned int i=0;

  if(stat(filename.c_str(),&buffer)==0)
  {
    // try to open the specified file
    try{
      std::auto_ptr<arm_motions_t> motions(arm_motions(filename.c_str(), xml_schema::flags::dont_validate));
      for(iterator=motions->arm_motion().begin(),i=0;iterator!=motions->arm_motion().end();iterator++,i++)
      {
        if(iterator->button_name()=="triangle")
        {
          new_motion.pan=iterator->pan();
          new_motion.tilt=iterator->tilt();
          new_motion.roll=iterator->roll();
          this->stored_poses[0]=new_motion;
        }
        else if(iterator->button_name()=="circle")
        {
          new_motion.pan=iterator->pan();
          new_motion.tilt=iterator->tilt();
          new_motion.roll=iterator->roll();
          this->stored_poses[1]=new_motion;
        }
        else if(iterator->button_name()=="cross")
        {
          new_motion.pan=iterator->pan();
          new_motion.tilt=iterator->tilt();
          new_motion.roll=iterator->roll();
          this->stored_poses[2]=new_motion;
        }
        else if(iterator->button_name()=="square")
        {
          new_motion.pan=iterator->pan();
          new_motion.tilt=iterator->tilt();
          new_motion.roll=iterator->roll();
          this->stored_poses[3]=new_motion;
        }
      }
      this->poses_file=filename;
    }catch (const xml_schema::exception& e){
      std::ostringstream os;
      os << e;
      /* handle exceptions */
      throw CException(_HERE_,os.str());
    }
  }
  else
    throw CException(_HERE_,"The motions file does not exist");
}

void CArmControl::save_poses(const std::string &filename)
{
  xml_schema::namespace_infomap map;
  arm_motions_t::arm_motion_sequence poses;
  arm_motions_t motions;

  try{
    std::ofstream output_file(filename.c_str(),std::ios_base::out);
    poses.push_back(arm_motion_t("triangle",this->stored_poses[0].pan,this->stored_poses[0].tilt,this->stored_poses[0].roll));
    poses.push_back(arm_motion_t("circle",this->stored_poses[1].pan,this->stored_poses[1].tilt,this->stored_poses[1].roll));
    poses.push_back(arm_motion_t("cross",this->stored_poses[2].pan,this->stored_poses[2].tilt,this->stored_poses[2].roll));
    poses.push_back(arm_motion_t("square",this->stored_poses[3].pan,this->stored_poses[3].tilt,this->stored_poses[3].roll));
    motions.arm_motion(poses);
    map[""].name="";
    map[""].schema="arm_motions_cfg.xsd";
    arm_motions(output_file,motions,map);
  }catch (const xml_schema::exception& e){
    std::ostringstream os;
    os << e;
    /* handle exceptions */
    throw CException(_HERE_,os.str());
  }
}

void CArmControl::update_pose(unsigned int pose_id)
{
  double pos;

  if(pose_id>3)
    throw CException(_HERE_,"Invalid pose ID");
  else
  {
    std::cout << "update pose" << std::endl;
    pos=this->motors[PAN]->get_current_angle();
    this->stored_poses[pose_id].pan=pos;
    pos=this->motors[TILT]->get_current_angle();
    this->stored_poses[pose_id].tilt=pos;
    pos=this->motors[ROLL]->get_current_angle();
    this->stored_poses[pose_id].roll=pos;
    if(this->poses_file!="")
      this->save_poses(this->poses_file);
    else
      this->save_poses("poses.xml");
  }
}

void CArmControl::execute_pose(unsigned int pose_id)
{
  double diff_pan, diff_tilt, diff_roll;
  double vel_pan, vel_tilt, vel_roll;
 
  if(pose_id>3)
    throw CException(_HERE_,"Invalid pose ID");
  else
  {
    std::cout << "execute pose" << std::endl;
    diff_pan=fabs(this->motors[PAN]->get_current_angle()-this->stored_poses[pose_id].pan);
    diff_tilt=fabs(this->motors[TILT]->get_current_angle()-this->stored_poses[pose_id].tilt);
    diff_roll=fabs(this->motors[ROLL]->get_current_angle()-this->stored_poses[pose_id].roll);
    if(diff_pan>diff_tilt)
    {
      if(diff_pan>diff_roll)
      {
        vel_pan=this->max_speed;
        vel_tilt=(diff_tilt/diff_pan)*this->max_speed;
        vel_roll=(diff_roll/diff_pan)*this->max_speed;
      }
      else
      {
        vel_tilt=this->max_speed;
        vel_pan=(diff_pan/diff_tilt)*this->max_speed;
        vel_roll=(diff_roll/diff_tilt)*this->max_speed;
      }
    }
    else
    {
      if(diff_tilt>diff_roll)
      {
        vel_tilt=this->max_speed;
        vel_pan=(diff_pan/diff_tilt)*this->max_speed;
        vel_roll=(diff_roll/diff_tilt)*this->max_speed;
      }
      else
      {
        vel_roll=this->max_speed;
        vel_pan=(diff_pan/diff_roll)*this->max_speed;
        vel_tilt=(diff_tilt/diff_roll)*this->max_speed;
      }
    }
    this->motors[PAN]->move_absolute_angle(this->stored_poses[pose_id].pan,vel_pan);
    this->motors[TILT]->move_absolute_angle(this->stored_poses[pose_id].tilt,vel_tilt);
    this->motors[ROLL]->move_absolute_angle(this->stored_poses[pose_id].roll,vel_roll);
  }
}

CArmControl::~CArmControl()
{
  // destroy thread and events
  if(this->thread_server->get_thread_state(this->position_monitor_thread_id)==starting ||
     this->thread_server->get_thread_state(this->position_monitor_thread_id)==active)
  {
    this->event_server->set_event(this->finish_thread_event_id);
    this->thread_server->end_thread(this->position_monitor_thread_id);
  }
  this->thread_server->detach_thread(this->position_monitor_thread_id);
  this->thread_server->delete_thread(this->position_monitor_thread_id);
  this->position_monitor_thread_id="";
  this->event_server->delete_event(this->finish_thread_event_id);
  this->finish_thread_event_id="";
  // destroy joystick object
  this->joy.close();
  // destroy motors
  if(this->motors.size()>0)
    if(this->motors[PAN]!=NULL)
    {
      delete this->motors[PAN];
      this->motors[PAN]=NULL;
    }
  if(this->motors.size()>1)
    if(this->motors[TILT]!=NULL)
    {
      delete this->motors[TILT];
      this->motors[TILT]=NULL;
    }
  if(this->motors.size()>2)
    if(this->motors[ROLL]!=NULL)
    {
      delete this->motors[ROLL];
      this->motors[ROLL]=NULL;
    }
}

